@@ -49,6 +49,7 @@ static volatile int child_wait_status;	/* Status from wait(). */
 
 RETSIGTYPE sigchld_handler(int sig)
 {
+  int save_errno = errno;
   int wait_pid;
   debug("Received SIGCHLD.");
   wait_pid = wait((int *)&child_wait_status);
@@ -62,6 +63,7 @@ RETSIGTYPE sigchld_handler(int sig)
 	child_terminated = 1;
     }
   signal(SIGCHLD, sigchld_handler);
+  errno = save_errno;
 }
 
 /* Process any buffered packets that have been received from the client. */
@@ -357,7 +359,7 @@ void process_output(fd_set *writeset)
 		  buffer_len(&stdin_buffer));
       if (len <= 0)
 	{
-	  shutdown(fdin, 1); /* We will no longer send. */
+	  shutdown(fdin, SHUT_WR); /* We will no longer send. */
 	  fdin = -1;
 	}
       else
@@ -477,7 +479,7 @@ void server_loop(int pid, int fdin_arg, int fdout_arg, int fderr_arg)
 	 cause a real eof by closing fdin. */
       if (stdin_eof && fdin != -1 && buffer_len(&stdin_buffer) == 0)
 	{
-	  shutdown(fdin, 1); /* We will no longer send. */
+	  shutdown(fdin, SHUT_WR); /* We will no longer send. */
 	  fdin = -1;
 	}
 
@@ -555,15 +557,15 @@ void server_loop(int pid, int fdin_arg, int fdout_arg, int fderr_arg)
 
   /* Close the file descriptors. */
   if (fdout != -1)
-    shutdown(fdout, 0);
+    shutdown(fdout, SHUT_RD);
   fdout = -1;
   fdout_eof = 1;
   if (fderr != -1)
-    shutdown(fderr, 0);
+    shutdown(fderr, SHUT_RD);
   fderr = -1;
   fderr_eof = 1;
   if (fdin != -1)
-    shutdown(fdin, 1);
+    shutdown(fdin, SHUT_WR);
   fdin = -1;
 
   /* Stop listening for channels; this removes unix domain sockets. */
